# AUTOGENERATED! DO NOT EDIT! File to edit: 04kraus.ipynb (unless otherwise specified).

__all__ = ['apply_kraus', 'partial_trace_kraus', 'povm_map']

# Cell
import numpy as np
import qutip as qt

# Cell
def apply_kraus(dm, kraus):
	r"""
	Applies a Kraus map to a density matrix $\rho$. The Kraus map consists in some number of operators
	satisfying $\sum_{i} \hat{K}_{i}^{\dagger}\hat{K}_{i} = \hat{I}$. $\rho$ is transformed via:

	$$\rho \rightarrow \sum_{i} \hat{K}_{i}\rho\hat{K}_{i}^{\dagger} $$
	"""
	return sum([kraus[j]*dm*kraus[j].dag() for j in range(len(kraus))])

# Cell
def partial_trace_kraus(keep, dims):
    r"""
    Constructs the Kraus map corresponding to the partial trace. Takes `keep` which is a single index or list of indices denoting
    subsystems to keep, and a list `dims` of dimensions of the overall tensor product Hilbert space.

    For illustration, to trace over the $i^{th}$ subsystem of $n$, one would construct Kraus operators:

    $$ \hat{K}_{i} = I^{\otimes i - 1} \otimes \langle i \mid \otimes I^{\otimes n - i}$$.
    """
    if type(keep) == int:
        keep = [keep]
    trace_over = [i for i in range(len(dims)) if i not in keep]
    indices = [{trace_over[0]:t} for t in range(dims[trace_over[0]])]
    for i in trace_over[1:]:
        new_indices = []
        for t in range(dims[i]):
            new_indices.extend([{**j, **{i: t}} for j in indices])
        indices = new_indices
    return [qt.tensor(*[qt.identity(d) if i in keep else qt.basis(d, index[i]).dag() for i, d in enumerate(dims)]) for index in indices]

# Cell
def povm_map(kraus, A, B=None):
    r"""
    Represents a Kraus map on Qbist probability vectors. Takes a list of Kraus operators, a POVM $A$ on the initial Hilbert space,
    and a POVM $B$ on the final Hilbert space. If $B$ isn't provided, it's assumed to be the same as $A$. Then the matrix elements of the map are:

    $$K_{j, i} = tr( \mathbb{K}(\frac{\hat{A}_{i}}{tr \hat{A}_{i}})\hat{B}_{j} ) $$

    Where $\mathbb{K}(\hat{O})$ denotes the Kraus map applied to $O$.
    """
    B = B if type(B) != type(None) else A
    return np.array([[(apply_kraus(a/a.tr(), kraus)*b).tr() for a in A] for b in B]).real


# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"clock": "01weylheisenberg.ipynb",
         "shift": "01weylheisenberg.ipynb",
         "displace": "01weylheisenberg.ipynb",
         "weyl_heisenberg_indices": "01weylheisenberg.ipynb",
         "displacement_operators": "01weylheisenberg.ipynb",
         "weyl_heisenberg_states": "01weylheisenberg.ipynb",
         "weyl_heisenberg_povm": "01weylheisenberg.ipynb",
         "to_weyl_heisenberg_basis": "01weylheisenberg.ipynb",
         "from_weyl_heisenberg_basis": "01weylheisenberg.ipynb",
         "load_fiducial": "02sics.ipynb",
         "sic_states": "02sics.ipynb",
         "sic_povm": "02sics.ipynb",
         "sic_gram": "02sics.ipynb",
         "hoggar_fiducial": "02sics.ipynb",
         "hoggar_indices": "02sics.ipynb",
         "hoggar_povm": "02sics.ipynb",
         "povm_gram": "03povm.ipynb",
         "povm_phi": "03povm.ipynb",
         "dm_probs": "03povm.ipynb",
         "probs_dm": "03povm.ipynb",
         "conditional_probs": "03povm.ipynb",
         "quantum_inner_product": "03povm.ipynb",
         "tensor_povm": "03povm.ipynb",
         "upgrade_povm": "03povm.ipynb",
         "apply_dims": "03povm.ipynb",
         "implement_povm": "03povm.ipynb",
         "discriminator_povm": "03povm.ipynb",
         "apply_kraus": "04kraus.ipynb",
         "partial_trace_kraus": "04kraus.ipynb",
         "povm_map": "04kraus.ipynb"}

modules = ["weyl_heisenberg.py",
           "sics.py",
           "povm.py",
           "kraus.py"]

doc_url = "https://heyredhat.github.io/qbism/qbism/"

git_url = "https://github.com/heyredhat/qbism/tree/master/"

def custom_doc_links(name): return None



# AUTOGENERATED! DO NOT EDIT! File to edit: 01weylheisenberg.ipynb (unless otherwise specified).

__all__ = ['clock', 'shift', 'displace', 'weyl_heisenberg_indices', 'displacement_operators', 'weyl_heisenberg_states',
           'weyl_heisenberg_povm', 'to_weyl_heisenberg_basis', 'from_weyl_heisenberg_basis']

# Cell
import numpy as np
import qutip as qt

# Cell
def clock(d):
    r"""
    The clock operator $\hat{Z}$ for dimension $d$.
    """
    w = np.exp(2*np.pi*1j/d)
    return qt.Qobj(np.diag([w**i for i in range(d)]))

# Cell
def shift(d):
    r"""
    The shift operator $\hat{X}$ for dimension $d$.
    """
    return sum([qt.basis(d, i+1)*qt.basis(d, i).dag()\
                    if i != d-1 else qt.basis(d, 0)*qt.basis(d, i).dag()\
                        for i in range(d) for j in range(d)])/d

# Cell
def displace(d, a, b):
    r"""
    The displacement operator $\hat{D}_{a,b} = (-e^{\frac{i\pi}{d}})^{ab}\hat{X}^{b}\hat{Z}^{a}$ for dimension $d$.
    """
    Z, X = clock(d), shift(d)
    return (-np.exp(1j*np.pi/d))**(a*b)*X**b*Z**a

# Cell
def weyl_heisenberg_indices(d):
    r"""
    Returns a list with entries $(a, b)$ for $a, b \in [0, d)$.
    """
    return [(a,b) for b in range(d) for a in range(d)]

# Cell
def displacement_operators(d):
    r"""
    Returns a dictionary associating $(a, b)$ with $\hat{D}_{a,b}$ for $a, b \in [0, d)$.
    """
    return dict([((a,b), displace(d, a, b)) for a, b in weyl_heisenberg_indices(d)])

# Cell
def weyl_heisenberg_states(fiducial):
    r"""
    Applies the $d^2$ displacement operators to a fiducial state, which can be either
    a ket or a density matrix.
    """
    d = fiducial.shape[0]
    D = displacement_operators(d)
    return [D[(a,b)]*fiducial if fiducial.type == "ket" else\
            D[(a,b)]*fiducial*D[(a,b)].dag()\
                for a, b in weyl_heisenberg_indices(d)]

# Cell
def weyl_heisenberg_povm(fiducial):
    r"""
    Generates a Weyl-Heisenberg POVM by applying the $d^2$ displacement operators to a
    fiducial state and then, if the fiducial state is a ket $\mid \psi \rangle$, forming the projector $\mid \psi \rangle \langle \psi \mid$, and normalizing by $\frac{1}{d}$.

    Note that if the fiducial state is a density matrix, it may be the case that it is invariant under some displacement operators, in which case you'll run into problems!
    """
    return [(1/fiducial.shape[0])*(state*state.dag() if state.type=='ket' else state) for state in weyl_heisenberg_states(fiducial)]

# Cell
def to_weyl_heisenberg_basis(O, D=None):
    r"""
    Expands a $d \times d$ operator $O$ in the Weyl-Heisenberg basis with components:

    $$ O_{a,b} = \frac{1}{d} tr ( \hat{D}_{a,b}^{\dagger} \hat{O} ) $$

    Returns a dictionary associating $(a, b)$ with components.
    """
    d = O.shape[0]
    D = D if type(D) != type(None) else displacement_operators(d)
    return dict([(index, (D_.dag()*O).tr()/d) for index, D_ in D.items()])

# Cell
def from_weyl_heisenberg_basis(C, D=None):
    r"""
    Given a dictionary of Weyl-Heisenberg components, returns the operator $O$
    in the standard basis:

    $$\hat{O} = \sum_{a,b} O_{a,b}\hat{D}_{a,b}$$
    """
    d = int(np.sqrt(len(C)))
    D = D if type(D) != type(None) else displacement_operators(d)
    return sum([coeff*D[index] for index, coeff in C.items()])


# AUTOGENERATED! DO NOT EDIT! File to edit: 02sics.ipynb (unless otherwise specified).

__all__ = ['load_fiducial', 'sic_states', 'sic_povm', 'sic_gram', 'hoggar_fiducial', 'hoggar_indices', 'hoggar_povm']

# Cell
import numpy as np
import qutip as qt
from itertools import product
import pkg_resources

from .weyl_heisenberg import *

# Cell
def load_fiducial(d):
    r"""
    Loads a Weyl-Heisenberg covariant SIC-POVM fiducial state of dimension $d$ from the repository provided here: http://www.physics.umb.edu/Research/QBism/solutions.html.
    """
    f = pkg_resources.resource_stream(__name__, "sic_povms/d%d.txt" % d)
    fiducial = []
    for line in f:
        if line.strip() != "":
            re, im = [float(v) for v in line.split()]
            fiducial.append(re + 1j*im)
    return qt.Qobj(np.array(fiducial)).unit()

# Cell
def sic_states(d):
    r"""
    Returns the $d^2$ states constructed by applying the Weyl-Heisenberg displacement operators to the SIC-POVM fiducial state of dimension $d$.
    """
    return weyl_heisenberg_states(load_fiducial(d))

# Cell
def sic_povm(d):
    r"""
    Returns a SIC-POVM of dimension $d$.
    """
    return weyl_heisenberg_povm(load_fiducial(d))

# Cell
def sic_gram(d):
    r"""
    The Gram matrix is the matrix of inner products: $G_{i,j} = \langle v_{i} \mid v_{j} \rangle$. For a SIC, this matrix should consist of 1's along the diagonal, and all other entries $\frac{1}{d+1}$:

    $$ \begin{pmatrix} 1 & \frac{1}{d+1} & \frac{1}{d+1} & \dots \\
                       \frac{1}{d+1} & 1 & \frac{1}{d+1} & \dots \\
                       \frac{1}{d+1} & \frac{1}{d+1} & 1 & \dots \\
                       \vdots & \vdots & \vdots & \ddots \end{pmatrix}$$

    """
    return np.array([[1 if i == j else 1/(d+1) for j in range(d**2)] for i in range(d**2)])

# Cell
def hoggar_fiducial():
    r"""
    Returns a fiducial state for the exceptional SIC in dimension $8$, the Hoggar SIC.

    Unnormalized: $\begin{pmatrix} -1 + 2i \\ 1 \\ 1 \\ 1 \\ 1 \\ 1 \\ 1 \\ 1 \end{pmatrix}$.
    """
    fiducial = qt.Qobj(np.array([-1 + 2j, 1, 1, 1, 1, 1, 1, 1])).unit()
    fiducial.dims = [[2,2,2],[1,1,1]]
    return fiducial

# Cell
def hoggar_indices():
    r"""
    Returns a list with entries $(a, b, c, d, e, f)$ for $a, b, c, d, e, f \in [0, 1]$.
    """
    return list(product([0,1], repeat=6))

# Cell
def hoggar_povm():
    r"""
    Constructs the Hoggar POVM, which is covariant under the tensor product of three copies of the $d=2$ Weyl-Heisenberg group. In other words, we apply the 64 displacement operators:

    $$ \hat{D}_{a, b, c, d, e, f} = X^{a}Z^{b} \otimes X^{c}Z^{d} \otimes X^{e}Z^{f} $$

    To the Hoggar fiducial state, form the corresponding projectors, and rescale by $\frac{1}{8}$.
    """
    Z, X = clock(2), shift(2)
    indices = hoggar_indices()
    D = dict([(I, qt.tensor(X**I[0]*Z**I[1],\
                            X**I[2]*Z**I[3],\
                            X**I[4]*Z**I[5])) for I in indices])
    fiducial = hoggar_fiducial()
    hoggar_states = [D[I]*fiducial for I in indices]
    return [(1/8)*state*state.dag() for state in hoggar_states]